@@ -28,7 +28,7 @@ class Optimizer:
     export_metrics=None, export_sysinfo=None,
     sampler=None, sampler_kwargs=None, pruner=None, pruner_kwargs=None,
     subprocess_kwargs=None, stdout_kwargs=None, stderr_kwargs=None,
-    params=None, objectives=None, attrs=None, lsep=None, gsep=None, secret=None):
+    params=None, objectives=None, attrs=None, lsep=None, gsep=None, secret=None, verbose=None):
     url = os.getenv('OPTUNA_URL') if url is None else url
     url = secret.get('optuna_url') if url is None else url
     self.url = url
@@ -67,11 +67,20 @@ class Optimizer:
     self.subprocess_kwargs = {} if subprocess_kwargs is None else subprocess_kwargs
     self.stdout_kwargs = {} if stdout_kwargs is None else stdout_kwargs
     self.stderr_kwargs = {} if stderr_kwargs is None else stderr_kwargs
+    if not isinstance(self.subprocess_kwargs, list):
+      self.subprocess_kwargs = [self.subprocess_kwargs]
+    if not isinstance(self.stdout_kwargs, list):
+      self.stdout_kwargs = [self.stdout_kwargs for _ in self.subprocess_kwargs]
+    if not isinstance(self.stderr_kwargs, list):
+      self.stderr_kwargs = [self.stderr_kwargs for _ in self.subprocess_kwargs]
     self.params = {} if params is None else params
     self.objectives = {} if objectives is None else objectives
     self.attrs = {} if attrs is None else attrs
     self.lsep = '/' if lsep is None else lsep
     self.gsep = '@' if gsep is None else gsep
+    if verbose is None:
+      verbose = int(os.getenv('OPTUNA_VERBOSE', 0))
+    self.verbose = verbose
 
   @staticmethod
   def initialize_sampler(sampler, sampler_kwargs):
@@ -246,7 +255,8 @@ class Optimizer:
            'processor': platform.processor(),
            'hostname': socket.gethostname(),
            'fqdn': socket.getfqdn(),
-           'ip': Optimizer.Logger.get_ip()}
+           'ip': Optimizer.Logger.get_ip(),
+           'cwd': os.getcwd()}
       return d
     
     @staticmethod
@@ -288,26 +298,32 @@ class Optimizer:
   
   def subprocess(self):
     subprocess_kwargs = copy.deepcopy(self.subprocess_kwargs)
-    stdout = self.subprocess_kwargs.get('stdout', None)
-    if stdout is not None:
-      if stdout in ['PIPE', 'STDOUT', 'DEVNULL']:
-        stdout = getattr(subprocess, stdout)
-      else:  # FILE
-        stdout = open(file=stdout, **self.stdout_kwargs)
-      subprocess_kwargs['stdout'] = stdout
-    stderr = subprocess_kwargs.get('stderr', None)
-    if stderr is not None:
-      if stderr in ['PIPE', 'STDOUT', 'DEVNULL']:
-        stderr = getattr(subprocess, stderr)
-      else:  # FILE
-        stderr = open(file=stderr, **self.stderr_kwargs)
-      subprocess_kwargs['stderr'] = stderr
-    result = subprocess.run(**subprocess_kwargs)
-    if isinstance(stdout, io.IOBase) and not stdout.closed:
-      stdout.close()
-    if isinstance(stderr, io.IOBase) and not stderr.closed:
-      stderr.close()
-    return result
+    stdout_kwargs = copy.deepcopy(self.stdout_kwargs)
+    stderr_kwargs = copy.deepcopy(self.stderr_kwargs)
+    results = []
+    for sub_kws, out_kws, err_kws in zip(
+      subprocess_kwargs, stdout_kwargs, stderr_kwargs):
+      stdout = sub_kws.get('stdout', None)
+      if stdout is not None:
+        if stdout in ['PIPE', 'STDOUT', 'DEVNULL']:
+          stdout = getattr(subprocess, stdout)
+        else:  # FILE
+          stdout = open(file=stdout, **out_kws)
+        sub_kws['stdout'] = stdout
+      stderr = sub_kws.get('stderr', None)
+      if stderr is not None:
+        if stderr in ['PIPE', 'STDOUT', 'DEVNULL']:
+          stderr = getattr(subprocess, stderr)
+        else:  # FILE
+          stderr = open(file=stderr, **err_kws)
+        sub_kws['stderr'] = stderr
+      result = subprocess.run(**sub_kws)
+      if isinstance(stdout, io.IOBase) and not stdout.closed:
+        stdout.close()
+      if isinstance(stderr, io.IOBase) and not stderr.closed:
+        stderr.close()
+      results.append(result)
+    return results
   
   def set_params(self, trial):
     for path, kwargs in self.params.items():
@@ -334,7 +350,8 @@ class Optimizer:
           if k.startswith(local_path):
             attrs[self.gsep.join([k, global_path])] = v
     print(f'Attributes: {len(attrs)}')
-    pprint(attrs)
+    if self.verbose:
+      pprint(attrs)
     for k, v in attrs.items():
       trial.set_user_attr(k, v)
   
@@ -352,7 +369,7 @@ class Optimizer:
   def objective(self, trial):
     self.set_params(trial)
     r = self.subprocess()
-    if r.returncode == 0:
+    if all([x.returncode == 0 for x in r]):
       self.set_attrs(trial)
       return self.get_objectives(trial)
     else:
@@ -388,5 +405,5 @@ class Optimizer:
     if self.export_csv is not None:
       df = study.trials_dataframe()
       p = Path(self.export_csv)
-      p.parent.mkdir(parents=True, exist_ok=False)
+      p.parent.mkdir(parents=True, exist_ok=True)
       df.to_csv(self.export_csv, index=False)
